/*
 *  Amiga Linux/m68k Loader
 *
 *   Copyright 1995 by Geert Uytterhoeven
 *
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file COPYING in the main directory of the Linux
 *  distribution for more details.
 */


%{

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>

#include "lilo_proto.h"


extern int conflex(void);
extern void conferror(const char *s);
extern int line;


struct Config Config = {
    NULL, NULL, NULL, {
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
    }, NULL, NULL
};


static struct BootRecord BootRecord;

static void ClearBootRecord()
{
    memset(&BootRecord, 0, sizeof(BootRecord));
}

static void Redefinition(const char *name)
{
    printf("%s:%d: Redefinition of `%s'\n", ConfigFile, line, name);
}

static const u_long *CopyLong(u_long data)
{
    u_long *p;

    if (!(p = malloc(sizeof(u_long))))
	Error_NoMemory();
    *p = data;
    return(p);
}


static const u_long *CopyLong2(u_long data1, u_long data2)
{
    u_long *p;

    if (!(p = malloc(2*sizeof(u_long))))
	Error_NoMemory();
    p[0] = data1;
    p[1] = data2;
    return(p);
}


static const u_short *CopyShort2(u_short data1, u_short data2)
{
    u_short *p;

    if (!(p = malloc(2*sizeof(u_short))))
	Error_NoMemory();
    p[0] = data1;
    p[1] = data2;
    return(p);
}

%}

%start file

%token HEADER ENDHEADER BOOTREC ENDREC FILEDEF BOOTDEV ALTDEV DEFAULT AUTO
       TIMEOUT AUX BAUD PASSWORD DEBUG PROMPT KICKRANGE TYPE IMAGE ARGS CHIPRAM
       FASTRAM MODEL ALIAS RAMDISK STRING NUMBER BOOLEAN BOOTOS AMIMODEL

%%

file	  : header bootrecs filedefs
	  ;

header	  : HEADER hdropts ENDHEADER
	    {
		if (!Config.BootDevice)
		    conferror("No boot device specified");
		ClearBootRecord();
	    }
	  ;

hdropts	  : /* empty */
	  | hdropts bootdev
	  | hdropts altdev
	  | hdropts default
	  | hdropts auto
	  | hdropts timeout
	  | hdropts aux
	  | hdropts baud
	  | hdropts mrpasswd
	  | hdropts debug
	  | hdropts prompt
	  | hdropts kickrange
	  ;

bootdev	  : BOOTDEV STRING
	    {
		if (Config.BootDevice)
		    Redefinition((char *)$1);
		Config.BootDevice = (char *)$2;
	    }
	  ;

altdev    : ALTDEV STRING NUMBER
	    {
		if (Config.AltDeviceName)
		    Redefinition((char *)$1);
		Config.AltDeviceName = (char *)$2;
		Config.AltDeviceUnit = CopyLong($3);
		if (strlen(Config.AltDeviceName) >= MAXALTDEVSIZE)
		    conferror("Alternate boot device name is too long");
	    }
	  ;

default   : DEFAULT STRING
	    {
		if (Config.Options.Default)
		    Redefinition((char *)$1);
		Config.Options.Default = (char *)$2;
	    }
	  ;

auto      : AUTO BOOLEAN
	    {
		if (Config.Options.Auto)
		    Redefinition((char *)$1);
		Config.Options.Auto = CopyLong($2);
	    }
	  ;

timeout   : TIMEOUT NUMBER
	    {
		if (Config.Options.TimeOut)
		    Redefinition((char *)$1);
		Config.Options.TimeOut = CopyLong($2);
	    }
	  ;

aux	  : AUX BOOLEAN
	    {
		if (Config.Options.Aux)
		    Redefinition((char *)$1);
		Config.Options.Aux = CopyLong($2);
	    }
	  ;

baud      : BAUD NUMBER
	    {
		if (Config.Options.Baud)
		    Redefinition((char *)$1);
		Config.Options.Baud = CopyLong($2);
	    }
	  ;

mrpasswd  : PASSWORD STRING
	    {
		if (Config.Options.MasterPassword)
		    Redefinition((char *)$1);
		Config.Options.MasterPassword = (char *)$2;
	    }
	  ;

debug     : DEBUG BOOLEAN
	    {
		if (Config.Options.Debug)
		    Redefinition((char *)$1);
		Config.Options.Debug = CopyLong($2);
	    }
	  ;

prompt	  : PROMPT STRING
	    {
		if (Config.Options.Prompt)
		    Redefinition((char *)$1);
		Config.Options.Prompt = (char *)$2;
	    }
	  ;

kickrange : KICKRANGE NUMBER NUMBER
	    {
		if (Config.Options.KickRange)
		    Redefinition((char *)$1);
		if ($2 > 65535 || $3 > 65535 || $3 < $2)
		    conferror("Invalid Kickstart version range");
		Config.Options.KickRange = CopyShort2($2, $3);
	    }
	  ;

bootrecs  : bootrec
	  | bootrecs bootrec
	  ;

bootrec   : BOOTREC STRING bootopts ENDREC
	    {
		BootRecord.Label = (char *)$2;
		if (BootRecord.OSType && *BootRecord.OSType == BOS_AMIGA) {
		    if (BootRecord.Kernel)
			conferror("You can't specify a kernel image for "
				  "AmigaOS");
		    if (BootRecord.Args)
			conferror("You can't specify arguments for AmigaOS");
		    if (BootRecord.Ramdisk)
			conferror("You can't specify a ramdisk image for "
				  "AmigaOS");
		} else if (!BootRecord.Kernel)
		    conferror("No kernel image specified");
		if (!AddBootRecord(&BootRecord))
		    conferror("Duplicate boot record label or alias");
		ClearBootRecord();
	    }
	  ;

bootopts  : /* empty */
	  | bootopts type
	  | bootopts image
	  | bootopts args
	  | bootopts passwd
	  | bootopts chipram
	  | bootopts fastram
	  | bootopts model
	  | bootopts modelnum
	  | bootopts alias
	  | bootopts ramdisk
	  ;

type      : TYPE BOOTOS
	    {
		if (BootRecord.OSType)
		    Redefinition((char *)$1);
		BootRecord.OSType = CopyLong($2);
	    }
	  ;

image     : IMAGE STRING
	    {
		if (BootRecord.Kernel)
		    Redefinition((char *)$1);
		BootRecord.Kernel = (char *)$2;
		AddFileDef(BootRecord.Kernel);
	    }
	  ;

args      : ARGS STRING
	    {
		if (BootRecord.Args)
		    Redefinition((char *)$1);
		BootRecord.Args = (char *)$2;
		if (strlen(BootRecord.Args) >= CL_SIZE)
		    conferror("Boot argument is too long");
	    }
	  ;

passwd    : PASSWORD STRING
	    {
		if (BootRecord.Password)
		    Redefinition((char *)$1);
		BootRecord.Password = (char *)$2;
	    }
	  ;

chipram   : CHIPRAM NUMBER
	    {
		if (BootRecord.ChipSize)
		    Redefinition((char *)$1);
		BootRecord.ChipSize = CopyLong($2);
	    }
	  ;

fastram   : FASTRAM NUMBER NUMBER
	    {
		int i;

		for (i = 0; i < arraysize(BootRecord.FastChunks); i++)
		    if (!BootRecord.FastChunks[i])
			break;
		if (i == arraysize(BootRecord.FastChunks))
		    conferror("Too many Fast RAM Chunks");
		BootRecord.FastChunks[i] = CopyLong2($2, $3);
	    }

model     : MODEL AMIMODEL
	    {
		if (BootRecord.Model)
		    Redefinition((char *)$1);
		BootRecord.Model = CopyLong($2);
	    }
	  ;

modelnum  : MODEL NUMBER
	    {
		if (BootRecord.Model)
		    Redefinition((char *)$1);
		BootRecord.Model = CopyLong($2);
	    }
	  ;

alias	  : ALIAS STRING
	    {
		if (BootRecord.Alias)
		    Redefinition((char *)$1);
		BootRecord.Alias = (char *)$2;
	    }
	  ;

ramdisk	  : RAMDISK STRING
	    {
		if (BootRecord.Ramdisk)
		    Redefinition((char *)$1);
		BootRecord.Ramdisk = (char *)$2;
		AddFileDef(BootRecord.Ramdisk);
	    }
	  ;

filedefs  : /* empty */
	  | filedefs filedef
	  ;

filedef	  : FILEDEF STRING
	    {
		AddFileDef((char *)$2);
	    }
	  ;

%%
