/*
 *  Amiga Linux/m68k Loader
 *
 *   Copyright 1995 by Geert Uytterhoeven
 *
 *
 *  The original idea for this program comes from the Linux distribution:
 *
 *	LILO - Generic Boot Loader for Linux ("LInux LOader")
 *	       by Werner Almesberger
 *
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file COPYING in the main directory of the Linux
 *  distribution for more details.
 */


#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/hdreg.h>

#include "lilo_proto.h"


const char LiloVersion[] = VERSION;


    /*
     *	Command Line Options
     */

static const char *ProgramName = NULL;
static int Verbose = 0;
static int Install = 0;
static int DoBackup = 0;
static int Uninstall = 0;
static int Force = 0;
static int DoChecksum = 0;
static const char *DosType = NULL;
static const char *SaveBootBlock = NULL;
static const char *RestoreBootBlock = NULL;
static const char *Device = NULL;
static const char *Root = NULL;


    /*
     *	Boot Code Definitions
     */

extern u_char _BootCodeStart[];
extern u_char _BootCodeEnd[];
extern u_char _LoaderVectorStart[];
extern u_char _LoaderVectorEnd[];
extern u_char _AltDeviceName[];
extern u_char _AltDeviceUnit[];


    /*
     *  Header for an AmigaOS Load Module
     */

extern u_char _HeadStart[];
extern u_char _HeadEnd[];


    /*
     *	Valid DosTypes
     */

struct DosTypeEntry {
    u_long ID;
    const char *ASCII;
    const char *Description;
};

static struct DosTypeEntry DosTypeEntries[] = {
    { ID_BOOT_DISK, "BOOU", "Generic boot disk" },
    { ID_OFS_DISK, "DOS0", "OFS disk" },
    { ID_FFS_DISK, "DOS1", "FFS disk" },
    { ID_OFS_INTL_DISK, "DOS2", "OFS INTL disk" },
    { ID_FFS_INTL_DISK, "DOS3", "FFS INTL disk" },
    { ID_OFS_DC_DISK, "DOS4", "OFS DC disk" },
    { ID_FFS_DC_DISK, "DOS5", "FFS DC disk" },
    { ID_MUFS_DISK, "muFS", "MultiUser FFS INTL disk" },
    { ID_MU_OFS_DISK, "muF0", "MultiUser OFS disk" },
    { ID_MU_FFS_DISK, "muF1", "MultiUser FFS disk" },
    { ID_MU_OFS_INTL_DISK, "muF2", "MultiUser OFS INTL disk" },
    { ID_MU_FFS_INTL_DISK, "muF3", "MultiUser FFS INTL disk" },
    { ID_MU_OFS_DC_DISK, "muF4", "MultiUser OFS DC disk" },
    { ID_MU_FFS_DC_DISK, "muF5", "MultiUser FFS DC disk" },
    { 0, NULL, NULL }
};


    /*
     *	Lilo Data
     */

static struct BootBlock BootBlock;

static int MapFileFH;
static const u_long *MapVector;
static int MapNumBlocks;

static char *LoaderData;
static int LoaderSize;
static const u_long *LoaderVector;
static int LoaderNumBlocks;
static u_long DosTypeID;

static dev_t BootDevice;


    /*
     *	Files
     */

const char *ConfigFile;
static const char *MapFile;
static const char *LoaderTemplate;
static const char *LoaderFile;
static const char *BackupFile;


    /*
     *	Function Prototypes
     */

int main(int argc, char *argv[]);

static void Error_Open(const char *name);
static void Error_OpenDir(const char *name);
static void Error_Read(const char *name);
static void Error_Write(const char *name);
static void Error_Seek(const char *name);
static void Error_Stat(const char *name);
static void Error_Ioctl(const char *name, const char *ioctl);
static int ReadFile(const char *name, void **data);
static void WriteFile(const char *name, const void *data, int size);
static void ReadBootBlock(const char *name);
static void WriteBootBlock(const char *name);
static void CreateBootBlock(void);
static void BackupBootBlock(void);
static void SetDosType(void);
static void FixChecksum(void);
static u_long CalcChecksum(void);
static void ReadConfigFile(void);
static int EqualStrings(const char *s1, const char *s2);
static void FixSpecialChars(char *s);
static void CreateMapFile(void);
static void WriteTagData(u_long tag, const void *data, u_long size);
static void WriteTagString(u_long tag, const char *s);
static const u_long *CreateVector(const char *name, int *numblocks);
static void Geometry(const char *name, dev_t *device, u_long *start);
static void PatchLoader(void);
static void WriteLoader(void);
static const char *CreateFileName(const char *name);
static const char *CreateBackupFileName(void);
static void Usage(void) __attribute__ ((noreturn));


    /*
     *	Print an Error Message and Exit
     */

void Die(const char *fmt, ...)
{
    va_list ap;

    fflush(stdout);
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    exit(1);
}


    /*
     *  Error Conditions
     */

void Error_NoMemory(void)
{
    Die("Can't allocate memory\n");
}

static void Error_Open(const char *name)
{
    Die("Can't open file `%s': %s\n", name, strerror(errno));
}

static void Error_OpenDir(const char *name)
{
    Die("Can't open directory `%s': %s\n", name, strerror(errno));
}

static void Error_Read(const char *name)
{
    Die("Can't read from file `%s': %s\n", name, strerror(errno));
}

static void Error_Write(const char *name)
{
    Die("Can't write to file `%s': %s\n", name, strerror(errno));
}

static void Error_Seek(const char *name)
{
    Die("Can't seek in file `%s': %s\n", name, strerror(errno));
}

static void Error_Stat(const char *name)
{
    Die("Can't get info about file `%s': %s\n", name, strerror(errno));
}

static void Error_Ioctl(const char *name, const char *ioctl)
{
    Die("Operation `%s' failed for file `%s': %s\n", ioctl, name,
	strerror(errno));
}


    /*
     *	Read a File
     */

static int ReadFile(const char *name, void **data)
{
    int fh, size = 0;

    if (Verbose)
	printf("Reading file `%s'\n", name);

    if ((fh = open(name, O_RDONLY)) == -1)
	Error_Open(name);
    if ((size = lseek(fh, 0, SEEK_END)) == -1)
	Error_Seek(name);
    if (!(*data = malloc(size)))
	Error_NoMemory();
    if (lseek(fh, 0, SEEK_SET) == -1)
	Error_Seek(name);
    if (read(fh, *data, size) != size)
	Error_Read(name);
    close(fh);
    return(size);
}


    /*
     *	Write a File
     */

static void WriteFile(const char *name, const void *data, int size)
{
    int fh;

    if (Verbose)
	printf("Writing file `%s'\n", name);

    if ((fh = open(name, O_CREAT|O_RDWR|O_TRUNC, S_IRUSR|S_IWUSR)) == -1)
	Error_Open(name);
    if (write(fh, data, size) != size)
	Error_Write(name);
    close(fh);
}


    /*
     *	Read a Boot Block from a Device or File
     */

static void ReadBootBlock(const char *name)
{
    int fh;

    if (Verbose)
	printf("Reading boot block from `%s'\n", name);

    if ((fh = open(name, O_RDONLY)) == -1)
	Error_Open(name);
    if (read(fh, &BootBlock, sizeof(BootBlock)) != sizeof(BootBlock))
	Error_Read(name);
    close(fh);
}


    /*
     *	Write a Boot Block to a Device
     */

static void WriteBootBlock(const char *name)
{
    int fh;

    if (Verbose)
	printf("Writing boot block to device `%s'\n", name);

    if ((fh = open(name, O_RDWR)) == -1)
	Error_Open(name);
    if (write(fh, &BootBlock, sizeof(BootBlock)) != sizeof(BootBlock))
	Error_Write(name);
    close(fh);
}


    /*
     *	Create the Boot Block
     */

static void CreateBootBlock(void)
{
    u_long maxsize, size;

    if (Verbose)
	puts("Creating boot block");

    size = _BootCodeEnd-_BootCodeStart;
    if (size > sizeof(BootBlock.Data))
	Die("%ld bytes too much boot code\n", size-sizeof(BootBlock.Data));
    if (Verbose)
	printf("%lu bytes boot code, %lu bytes free\n", size,
	       sizeof(BootBlock.Data)-size);
    BootBlock.ID = DosTypeID;
    if (BootBlock.ID == ID_BOOT_DISK)
	BootBlock.DosBlock = 0;
    memcpy(BootBlock.Data, _BootCodeStart, size);

    maxsize = (_LoaderVectorEnd-_LoaderVectorStart)/sizeof(u_long);
    if (LoaderNumBlocks > maxsize-1)
	Die("Loader vector is too large for boot block (%ld entries too "
	    "much)\n",
	    LoaderNumBlocks-maxsize-1);
    if (Verbose)
	printf("%d entries for loader vector blocks, %lu entries free\n",
	       LoaderNumBlocks, maxsize-1-LoaderNumBlocks);
    memcpy((void *)((u_long)BootBlock.Data+(u_long)_LoaderVectorStart-
		    (u_long)_BootCodeStart), LoaderVector,
	   (LoaderNumBlocks+1)*sizeof(u_long));
    if (Config.AltDeviceName) {
	strcpy((void *)((u_long)BootBlock.Data+(u_long)_AltDeviceName-
			(u_long)_BootCodeStart), Config.AltDeviceName);
	memcpy((void *)((u_long)BootBlock.Data+(u_long)_AltDeviceUnit-
			(u_long)_BootCodeStart), &Config.AltDeviceUnit,
			     sizeof(Config.AltDeviceUnit));
    }
    BootBlock.LiloID = LILO_ID;
    FixChecksum();
}


    /*
     *  Backup the Boot Block if it wasn't backed up before
     */

static void BackupBootBlock(void)
{
    struct stat info;

    if (stat(BackupFile, &info) == -1 || DoBackup) {
	WriteFile(BackupFile, &BootBlock, sizeof(BootBlock));
	printf("Boot block backed up to file `%s'\n", BackupFile);
    } else if (Verbose)
        printf("Boot block backup `%s' already exists. No backup taken.\n",
	       BackupFile);
}


    /*
     *	Set the DosType of the Boot Block
     */

static void SetDosType(void)
{
    int i;

    for (i = 0; DosTypeEntries[i].ID; i++)
	if (!strcmp(DosType, DosTypeEntries[i].ASCII))
	    break;
    if (DosTypeEntries[i].ID)
	DosTypeID = DosTypeEntries[i].ID;
    else {
	printf("Unknown DosType `%s'\n", DosType);
	puts("Valid DosTypes are:");
	for (i = 0; DosTypeEntries[i].ID; i++)
	    printf(" %s", DosTypeEntries[i].ASCII);
	Die("\n");
    }
    BootBlock.ID = DosTypeID;
    if (Verbose)
	printf("DosType set to `%s'\n", DosType);
    FixChecksum();
}


    /*
     *	Fix the Checksum of the Boot Block
     */

static void FixChecksum(void)
{
    if (Verbose)
	puts("Calculating boot block checksum");

    BootBlock.Checksum = 0;
    BootBlock.Checksum = CalcChecksum();
}


    /*
     *	Calculate the Boot Block Checksum
     */

static u_long CalcChecksum(void)
{
    u_long _res;
    struct BootBlock *bb = &BootBlock;

    __asm __volatile ("1: addl %0@+,%1;"
		      "   addxl %5,%1;"
		      "   dbf %4,1b;"
		      : "=a" (bb), "=d" (_res)
		      : "0" (bb), "1" (0), "d" (255), "d" (0));
    return(~_res);
}


    /*
     *	Read the Configuration File
     */

static void ReadConfigFile(void)
{
    if (Verbose)
	printf("Reading configuration file `%s'\n", ConfigFile);

    if (!(confin = fopen(ConfigFile, "r")))
	Error_Open(ConfigFile);
    confparse();
    fclose(confin);

    Device = Config.BootDevice;
    if (!Config.Options.Default)
	Config.Options.Default = Config.Records->Label;
    if (Config.Options.Prompt)
	FixSpecialChars((char *)Config.Options.Prompt);
}


    /*
     *	Check whether two strings exist and are equal
     */

static int EqualStrings(const char *s1, const char *s2)
{
    return(s1 && s2 && !strcmp(s1, s2) ? 1 : 0);
}


    /*
     *  Interprete the special characters in a string
     */

static void FixSpecialChars(char *s)
{
    char *d = s;

    while (*s)
	if (s[0] == '\\' && s[1])
	    switch (s[1]) {
		case 'b':
		    *d++ = '\b';
		    s += 2;
		    break;
		case 'e':
		    *d++ = '\e';
		    s += 2;
		    break;
		case 'n':
		    *d++ = '\n';
		    s += 2;
		    break;
		case 'r':
		    *d++ = '\r';
		    s += 2;
		    break;
		case 't':
		    *d++ = '\t';
		    s += 2;
		    break;
		case '\\':
		    *d++ = '\\';
		    s += 2;
		    break;
		default:
		    *d++ = *s++;
		    break;
	    }
	else
	    *d++ = *s++;
    *d = '\0';
}


    /*
     *	Add a New Boot Record
     */

int AddBootRecord(const struct BootRecord *record)
{
    struct BootRecord **p, *new;

    for (p = &Config.Records; *p; p = &(*p)->Next)
	if (EqualStrings((*p)->Label, record->Label) ||
	    EqualStrings((*p)->Alias, record->Label) ||
	    EqualStrings((*p)->Label, record->Alias) ||
	    EqualStrings((*p)->Alias, record->Alias))
	    return(0);
    if (!(new = malloc(sizeof(struct BootRecord))))
	Error_NoMemory();
    *new = *record;
    *p = new;
    return(1);
}


    /*
     *	Add a New File Definition
     */

void AddFileDef(const char *path)
{
    struct FileDef **p, *new;

    for (p = &Config.Files; *p; p = &(*p)->Next)
	if (EqualStrings((*p)->Path, path))
	    return;
    if (!(new = malloc(sizeof(struct FileDef))))
	Error_NoMemory();
    new->Next = NULL;
    new->Path = path;
    new->Vector = NULL;
    *p = new;
}


    /*
     *	Create the Map File
     */

static void CreateMapFile(void)
{
    const struct BootRecord *record, *defrecord = NULL;
    struct FileDef *file;
    int i, vecsize, numblocks;

    if (Verbose)
	printf("Creating map file `%s'\n", MapFile);

    if ((MapFileFH = open(MapFile, O_CREAT|O_RDWR|O_TRUNC, S_IRUSR|S_IWUSR)) ==
	-1)
	Error_Open(MapFile);

    /* File Identification */
    WriteTagData(TAG_LILO, NULL, 0);

    /* Boot Options */
    if (Verbose)
	puts("  [ Boot Options ]");
    WriteTagData(TAG_HEADER, NULL, 0);
    if (Config.Options.Default)
	WriteTagString(TAG_DEFAULT, Config.Options.Default);
    if (Config.Options.Auto)
	WriteTagData(TAG_AUTO, Config.Options.Auto, 4);
    if (Config.Options.TimeOut)
	WriteTagData(TAG_TIME_OUT, Config.Options.TimeOut, 4);
    if (Config.Options.Aux)
	WriteTagData(TAG_AUX, Config.Options.Aux, 4);
    if (Config.Options.Baud)
	WriteTagData(TAG_BAUD, Config.Options.Baud, 4);
    if (Config.Options.MasterPassword)
	WriteTagString(TAG_MASTER_PASSWORD, Config.Options.MasterPassword);
    if (Config.Options.Debug)
	WriteTagData(TAG_DEBUG, Config.Options.Debug, 4);
    if (Config.Options.Prompt)
	WriteTagString(TAG_PROMPT, Config.Options.Prompt);
    if (Config.Options.KickRange)
	WriteTagData(TAG_KICKRANGE, Config.Options.KickRange, 4);
    WriteTagData(TAG_HEADER_END, NULL, 0);

    /* Boot Records */
    if (Verbose)
	puts("  [ Boot Records ]");
    else
	puts("\nBoot records:");
    for (record = Config.Records; record; record = record->Next) {
	if (EqualStrings(Config.Options.Default, record->Label) ||
	    EqualStrings(Config.Options.Default, record->Alias)) {
	    defrecord = record;
	    if (record->Password)
		Die("Default boot record `%s' must not have a password\n",
		    record->Label);
	}
	if (defrecord == record)
	    printf("\e[1m");
	printf("    %s\n", record->Label);
	if (defrecord == record)
	    printf("\e[0m");

	WriteTagString(TAG_BOOT_RECORD, record->Label);
	if (record->Alias)
	    WriteTagString(TAG_ALIAS, record->Alias);
	if (record->OSType)
	    WriteTagData(TAG_OS_TYPE, record->OSType, 4);
	if (record->Kernel)
	    WriteTagString(TAG_KERNEL, record->Kernel);
	if (record->Args)
	    WriteTagString(TAG_ARGUMENTS, record->Args);
	if (record->Password)
	    WriteTagString(TAG_PASSWORD, record->Password);
	if (record->ChipSize)
	    WriteTagData(TAG_CHIP_RAM_SIZE, record->ChipSize, 4);
	for (i = 0; i < arraysize(record->FastChunks); i++)
	    if (record->FastChunks[i])
		WriteTagData(TAG_FAST_RAM_CHUNK, record->FastChunks[i], 8);
	if (record->Model)
	    WriteTagData(TAG_MODEL, record->Model, 4);
	if (record->Ramdisk)
	    WriteTagString(TAG_RAMDISK, record->Ramdisk);
	WriteTagData(TAG_BOOT_RECORD_END, NULL, 0);
    }
    if (!defrecord)
	Die("Default boot record `%s' does not exist\n",
	    Config.Options.Default);

    /* File Definitions */
    if (Verbose)
	puts("  [ File Definitions ]");
    for (file = Config.Files; file; file = file->Next) {
	vecsize = 0;
	if ((file->Vector = CreateVector(file->Path, &numblocks)))
	    vecsize = (numblocks+1)*sizeof(u_long);
	else if ((!defrecord->OSType || *defrecord->OSType != BOS_AMIGA) &&
		 EqualStrings(defrecord->Kernel, file->Path))
	    Die("Kernel image `%s' for default boot record `%s' does not "
		"exist\n",
		file->Path, defrecord->Label);
	else
	    printf("Warning: file `%s' doesn't exist\n", file->Path);
	if (Verbose && file->Vector)
	    printf("    %s (%ld bytes)\n", file->Path, file->Vector[0]);
	WriteTagString(TAG_FILE_DEF, file->Path);
	if (file->Vector)
	    WriteTagData(TAG_VECTOR, file->Vector, vecsize);
	WriteTagData(TAG_FILE_DEF_END, NULL, 0);
    }

    /* End of File */
    WriteTagData(TAG_EOF, NULL, 0);
    close(MapFileFH);

    if (!Verbose)
	puts("");
}


    /*
     *	Write a Tag Record to the Map File
     */

static void WriteTagData(u_long tag, const void *data, u_long size)
{
    struct TagRecord tr;

    tr.Tag = tag;
    tr.Size = size;
    if (write(MapFileFH, &tr, sizeof(tr)) != sizeof(tr))
	Error_Write(MapFile);
    if (size) {
	if (write(MapFileFH, data, size) != size)
	    Error_Write(MapFile);
    }
}


static void WriteTagString(u_long tag, const char *s)
{
    WriteTagData(tag, s, (strlen(s)+4)&~3);
}


    /*
     *	Create a Vector for a File
     */

static const u_long *CreateVector(const char *name, int *numblocks)
{
    int fh, size, i;
    u_long *vector = NULL;
    u_long offset, start, blksize;
    dev_t device;
    int relblksize;
    static u_int first = 1;
    static dev_t altdev;

    if (!name || ((fh = open(name, O_RDONLY)) == -1))
	return(NULL);
    if ((size = lseek(fh, 0, SEEK_END)) == -1)
	Error_Seek(name);
    *numblocks = (size+LILO_BLOCKSIZE-1)/LILO_BLOCKSIZE;
    if (!(vector = malloc((*numblocks+1)*sizeof(u_long))))
	Error_NoMemory();
    vector[0] = size;

    Geometry(name, &device, &start);

    if (!Config.AltDeviceName) {
	if (MAJOR(device) != MAJOR(BootDevice))
	    Die("File `%s' must reside on physical device `%s'\n", name,
		Device);
	switch (MAJOR(BootDevice)) {
	    case FLOPPY_MAJOR:
		if (MINOR(device) != MINOR(BootDevice))
		    Die("File `%s' must reside on physical floppy device "
			"`%s'\n",
			name, Device);
		break;
	    case HD_MAJOR:
		if ((MINOR(device) & 0xc0) != (MINOR(BootDevice) & 0xc0))
		    Die("File `%s' must reside on physical IDE device `%s'\n",
			name, Device);
		break;
	    case SCSI_DISK_MAJOR:
		if ((MINOR(device) & 0xf0) != (MINOR(BootDevice) & 0xf0))
		    Die("File `%s' must reside on physical SCSI device `%s'\n",
			name, Device);
		break;
	    default:
		Die("Major device %d not supported (yet)\n", MAJOR(device));
		break;
	}
    } else if (first) {
	altdev = device;
	printf("Assuming `%s' resides on `%s:%ld'\n", name,
	       Config.AltDeviceName, *Config.AltDeviceUnit);
	first = 0;
    } else {
	if (MAJOR(device) != MAJOR(altdev))
	    Die("File `%s' must reside on `%s:%ld'\n", name,
		Config.AltDeviceName, *Config.AltDeviceUnit);
	switch (MAJOR(altdev)) {
	    case FLOPPY_MAJOR:
		if (MINOR(device) != MINOR(altdev))
		    Die("File `%s' must reside on physical floppy device "
			"`%s:%ld'\n",
			name, Config.AltDeviceName, *Config.AltDeviceUnit);
		break;
	    case HD_MAJOR:
		if ((MINOR(device) & 0xc0) != (MINOR(altdev) & 0xc0))
		    Die("File `%s' must reside on physical IDE device "
			"`%s:%ld'\n",
			name, Config.AltDeviceName, *Config.AltDeviceUnit);
		break;
	    case SCSI_DISK_MAJOR:
		if ((MINOR(device) & 0xf0) != (MINOR(altdev) & 0xf0))
		    Die("File `%s' must reside on physical SCSI device "
			"`%s:%ld'\n",
			name, Config.AltDeviceName, *Config.AltDeviceUnit);
		break;
	    default:
		Die("Major device %d not supported (yet)\n", MAJOR(device));
		break;
	}
    }

    if (ioctl(fh, FIGETBSZ, &blksize) == -1)
	Error_Ioctl(name, "FIGETBSZ");
    if (blksize % LILO_BLOCKSIZE)
	Die("File `%s' resides on a device with an unsupported block size "
	    "(%ld bytes)\n", name, blksize);
    relblksize = blksize/LILO_BLOCKSIZE;

    for (i = 0; i < *numblocks; i++) {
	if (!(i % relblksize)) {
	    offset = i/relblksize;
	    if (ioctl(fh, FIBMAP, &offset) == -1)
		Error_Ioctl(name, "FIBMAP");
	    if (offset)
		offset = start+offset*blksize;
	} else if ((offset = vector[i]))
	    offset += LILO_BLOCKSIZE;
	vector[i+1] = offset;
    }
    close(fh);
    return(vector);
}


    /*
     *	Get the Geometry for the Partition a File belongs to
     */

static void Geometry(const char *name, dev_t *device, u_long *start)
{
    struct hd_geometry geo;
    char devname[PATH_MAX+1];
    struct stat info;
    DIR *dir;
    const struct dirent *dirent;
    int fh, found = 0;

    if (stat(name, &info) == -1)
	Error_Stat(name);
    if (!S_ISREG(info.st_mode))
	Die("%s is not a regular file\n", name);
    *device = info.st_dev;

    if (!(dir = opendir("/dev")))
	Error_OpenDir("/dev");
    while ((dirent = readdir(dir))) {
	sprintf(devname, "/dev/%s", dirent->d_name);
	if ((stat(devname, &info) != -1) && S_ISBLK(info.st_mode) &&
	    (info.st_rdev == *device)) {
	    if ((fh = open(devname, O_RDONLY)) == -1)
		Error_Open(devname);
	    if (ioctl(fh, HDIO_GETGEO, &geo) == -1)
		Error_Ioctl(devname, "HDIO_GETGEO");
	    close(fh);
	    found = 1;
	    break;
	}
    }
    closedir(dir);
    if (found)
	*start = 512*geo.start;
    else
	Die("Can't find special device entry for file `%s'\n", name);
}


    /*
     *	Patch the Loader with the Map Vector
     */

static void PatchLoader(void)
{
    int i;
    const u_long pattern[] = {
	LILO_ID, LILO_MAPVECTORID
    };
    u_long *data = NULL;
    u_long maxsize;

    for (i = 0; i < LoaderSize-sizeof(pattern)+1; i++)
	if (!strncmp(&LoaderData[i], (char *)pattern, sizeof(pattern))) {
	    data = (u_long *)&LoaderData[i];
	    break;
	}
    if (!data)
	Die("Couldn't find magic key in loader template\n");
    maxsize = data[2];
    if (MapNumBlocks > maxsize-1)
	Die("Map vector is too large for loader (%ld entries too much)\n",
	    MapNumBlocks-maxsize-1);
    if (Verbose)
	printf("%d entries for map vector blocks, %lu entries free\n",
	       MapNumBlocks, maxsize-1-MapNumBlocks);
    memcpy(&data[3], MapVector, (MapNumBlocks+1)*sizeof(u_long));
}


    /*
     *  Write the Loader (incl. Header Code)
     */

static void WriteLoader(void)
{
    int fh;
    u_long headsize = _HeadEnd-_HeadStart;

    if (Verbose)
	printf("Writing loader to file `%s'\n", LoaderFile);

    if ((fh = open(LoaderFile, O_CREAT|O_RDWR|O_TRUNC, S_IRUSR|S_IWUSR)) == -1)
	Error_Open(LoaderFile);
    if (write(fh, _HeadStart, headsize) != headsize)
	Error_Write(LoaderFile);
    if (write(fh, LoaderData, LoaderSize) != LoaderSize)
	Error_Write(LoaderFile);
    close(fh);
}


    /*
     *	Create a Complete File Name
     */

static const char *CreateFileName(const char *name)
{
    int size;
    char *s;

    size = strlen(name)+1;
    if (Root)
	size += strlen(Root);
    if (!(s = malloc(size)))
	Error_NoMemory();
    if (Root) {
	strcpy(s, Root);
	strcat(s, name);
    } else
	strcpy(s, name);
    return(s);
}


    /*
     *  Create the File Name for the Boot Block Backup
     */

const char *CreateBackupFileName(void)
{
    int size;
    char *s;

    size = strlen(LILO_BACKUPFILEPREFIX)+7;
    if (Root)
	size += strlen(Root);
    if (!(s = malloc(size)))
	Error_NoMemory();
    sprintf(s, "%s%s.%02x.%02x", Root ? Root : "", LILO_BACKUPFILEPREFIX,
	    MAJOR(BootDevice), MINOR(BootDevice));
    return(s);
}


    /*
     *	Print the Usage Template and Exit
     */

static void Usage(void)
{
    Die("%s\nUsage: %s [options]\n\n"
	"Valid options are:\n"
	"    -h, --help                Display this usage information\n"
	"    -v, --verbose             Enable verbose mode\n"
	"    -f, --force               Force installation over an unknown boot block\n"
	"    -b, --backup              Force a boot block backup on installation\n"
	"    -u, --uninstall           Uninstall the boot block\n"
	"    -w, --root prefix         Use another root directory\n"
	"    -r, --restore-from file   File to restore the boot block from\n"
	"    -s, --save-to file        File to save the boot block to\n"
	"    -d, --device device       Override the block special device\n"
	"    -t, --dostype dostype     Dostype for the boot block\n"
	"    -c, --checksum            Fix the boot block checksum\n"
	"    -V, --version             Display version information\n",
	LiloVersion, ProgramName);
}


    /*
     *	Main Routine
     */

int main(int argc, char *argv[])
{
    struct stat info;
    int i;
    const char *device = NULL;

    ProgramName = argv[0];

    while (--argc > 0) {
	argv++;
	if (!strcmp(argv[0], "-h") || !strcmp(argv[0], "--help"))
	    Usage();
	else if (!strcmp(argv[0], "-v") || !strcmp(argv[0], "--verbose"))
	    Verbose = 1;
	else if (!strcmp(argv[0], "-f") || !strcmp(argv[0], "--force"))
	    Force = 1;
	else if (!strcmp(argv[0], "-b") || !strcmp(argv[0], "--backup"))
	    DoBackup = 1;
	else if (!strcmp(argv[0], "-u") || !strcmp(argv[0], "--uninstall"))
	    Uninstall = 1;
	else if (!strcmp(argv[0], "-c") || !strcmp(argv[0], "--checksum"))
	    DoChecksum = 1;
	else if (!strcmp(argv[0], "-V") || !strcmp(argv[0], "--version"))
	    Die(LiloVersion);
	else if (!strcmp(argv[0], "-w") || !strcmp(argv[0], "--root"))
	    if (argc-- > 1 && !Root) {
		Root = argv[1];
		argv++;
	    } else
		Usage();
	else if (!strcmp(argv[0], "-r") || !strcmp(argv[0], "--restore-from"))
	    if (argc-- > 1 && !RestoreBootBlock) {
		RestoreBootBlock = argv[1];
		argv++;
	    } else
		Usage();
	else if (!strcmp(argv[0], "-s") || !strcmp(argv[0], "--save-to"))
	    if (argc-- > 1 && !SaveBootBlock) {
		SaveBootBlock = argv[1];
		argv++;
	    } else
		Usage();
	else if (!strcmp(argv[0], "-d") || !strcmp(argv[0], "--device"))
	    if (argc-- > 1 && !device) {
		device = argv[1];
		argv++;
	    } else
		Usage();
	else if (!strcmp(argv[0], "-t") || !strcmp(argv[0], "--dostype"))
	    if (argc-- > 1 && !DosType) {
		DosType = argv[1];
		argv++;
	    } else
		Usage();
	else
	    Usage();
    }
    if (!Uninstall && !RestoreBootBlock && !SaveBootBlock && !DoChecksum)
        Install = 1;
    else if (DoBackup)
	Die("You can't specify --backup with --uninstall, --restore-from, "
	    "--save-to or \n--checksum\n");
    if (RestoreBootBlock && Uninstall)
	Die("You can't specify both --restore-from and --uninstall\n");

    if (Verbose)
	printf(LiloVersion);

    ConfigFile = CreateFileName(LILO_CONFIGFILE);
    MapFile = CreateFileName(LILO_MAPFILE);
    LoaderTemplate = CreateFileName(LILO_LOADERTEMPLATE);
    LoaderFile = CreateFileName(LILO_LOADERFILE);

    ReadConfigFile();
    if (device) {
        if (Verbose)
            printf("Using specified device `%s'\n", device);
        Device = device;
    }

    if (stat(Device, &info) == -1)
	Error_Stat(Device);
    if (!S_ISBLK(info.st_mode))
	Die("%s is not a block special device\n", Device);
    BootDevice = info.st_rdev;
    BackupFile = CreateBackupFileName();

    ReadBootBlock(Device);
    printf("%s: ", Device);
    DosTypeID = BootBlock.ID;
    for (i = 0; DosTypeEntries[i].ID; i++)
	if (BootBlock.ID == DosTypeEntries[i].ID)
	    break;
    if (DosTypeEntries[i].ID)
	printf("%s", DosTypeEntries[i].Description);
    else {
	printf("Type 0x%08lx", BootBlock.ID);
	if (!Force) {
	    printf("\n");
	    Die("Use option --force to force continue\n");
	}
	DosTypeID = ID_BOOT_DISK;
    }
    printf("%s, Boot block checksum is %s\n",
	   BootBlock.LiloID == LILO_ID ? " (created by Lilo)" : "",
	   CalcChecksum() ? "Incorrect" : "OK");

    if (SaveBootBlock) {
	WriteFile(SaveBootBlock, &BootBlock, sizeof(BootBlock));
	printf("Boot block saved to file `%s'\n", SaveBootBlock);
    }
    if (RestoreBootBlock) {
	ReadBootBlock(RestoreBootBlock);
	printf("Boot block read from file `%s'\n", RestoreBootBlock);
    }
    if (Uninstall) {
	ReadBootBlock(BackupFile);
	printf("Original boot block read from file `%s'\n", BackupFile);
    }
    if (DosType)
	SetDosType();
    if (Install) {
	BackupBootBlock();
	CreateMapFile();
	LoaderSize = ReadFile(LoaderTemplate, (void **)&LoaderData);
	if (!(MapVector = CreateVector(MapFile, &MapNumBlocks)))
	    Error_Open(MapFile);
	PatchLoader();
	WriteLoader();
	if (!(LoaderVector = CreateVector(LoaderFile, &LoaderNumBlocks)))
	    Error_Open(LoaderFile);
	CreateBootBlock();
    }
    if (DoChecksum)
	FixChecksum();
    if (RestoreBootBlock || Install || Uninstall || DoChecksum)
	WriteBootBlock(Device);
    exit(0);
}
