/*
 *  Amiga Linux/m68k Loader
 *
 *   Copyright 1995 by Geert Uytterhoeven
 *
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file COPYING in the main directory of the Linux
 *  distribution for more details.
 */


#define V36_COMPATIBILITY	/* Make it work with the A3000 Beta ROM */


#include <stddef.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>

#include <linux/linkage.h>
#include <asm/amigahw.h>

#include "bootlib.h"
#include "lilo.h"
#include "loader_proto.h"


    /*
     *  Program Entry
     *
     *  Pass the Pointer to the LiloTable (in d0) to Main
     */

asm(".text\n"
ALIGN_STR "\n"
".globl " SYMBOL_NAME_STR(_start) ";\n"
SYMBOL_NAME_STR(_start) ":
	movel	d0,sp@-
	jbsr	(" SYMBOL_NAME_STR(Main) ")
	addql	#4,sp
	rts
");


const char LiloVersion[] = VERSION;


    /*
     *	Lilo Alert
     */

#define AN_LILO		(AT_DeadEnd|0x4c000000)	/* `L' */
#define AO_LiloMap	(0x004d4150)		/* `MAP': Corrupt Map File */


    /*
     *	Console Echo Modes
     */

#define ECHO_NONE	(0)
#define ECHO_NORMAL	(1)
#define ECHO_DOT	(2)


    /*
     *	Custom Hardware Bases
     *
     *	Make sure these are the physical and not the virtual addresses
     */

#undef custom
#define custom		((*(volatile struct CUSTOM *)(CUSTOM_PHYSADDR)))
#undef ciaa
#define ciaa		((*(volatile struct CIA *)(CIAA_PHYSADDR)))


    /*
     *	Boot Library Functions
     */

static const struct LiloTable *LiloTable;


    /*
     *	Library Bases
     */

#ifdef V36_COMPATIBILITY
#define LIB_VERS	(36)
#else /* V36_COMPATIBILITY */
#define LIB_VERS	(37)
#endif /* V36_COMPATIBILITY */

const struct ExecBase *SysBase = NULL;
const struct ExpansionBase *ExpansionBase = NULL;
const struct GfxBase *GfxBase = NULL;
static const struct IntuitionBase *IntuitionBase = NULL;
static const struct InputBase *InputBase = NULL;
#ifdef V36_COMPATIBILITY
static const struct KeyboardBase *KeyboardBase = NULL;
#endif /* V36_COMPATIBILITY */


    /*
     *	Console Stuff
     */

static u_char Title[] = "Amiga Linux/m68k Loader";

static struct NewScreen NewScreen = {
    0, 0, 0, 0, 2, 0, 1, HIRES, CUSTOMSCREEN, NULL, NULL, NULL, NULL
};
static struct NewWindow NewWindow = {
    0, 0, 0, 0, 2, 3, 0, WFLG_SIMPLE_REFRESH|WFLG_ACTIVATE, NULL, NULL,
    NULL, NULL, NULL, 0, 0, 0, 0, CUSTOMSCREEN
};
static struct Screen *Screen;
static struct Window *Window;
static struct MsgPort *ConsolePort;
static struct IOStdReq *ConsoleRequest = NULL;
static struct MsgPort *TimerPort;
static struct timerequest *TimerRequest;
static struct MsgPort *InputPort;
static struct IOStdReq *InputRequest;
#ifdef V36_COMPATIBILITY
static struct MsgPort *KeyboardPort;
static struct IOStdReq *KeyboardRequest;
static u_char KeyboardMatrix[13];
#endif /* V36_COMPATIBILITY */
static u_long TimeOutSignalMask = 0;
static char FmtBuffer[256], InBuffer[CL_SIZE];
static u_long ConsoleOpened = FALSE;
static void *OldDebug;


    /*
     *	Map Data
     */

#define MAXVECTORSIZE	(1024)		/* Should be sufficient */

struct FileVectorData {
    u_long ID[2];
    u_long MaxVectorSize;
    u_long Vector[MAXVECTORSIZE];
};

static const struct FileVectorData MapVectorData = {
    { LILO_ID, LILO_MAPVECTORID }, MAXVECTORSIZE
};

static u_char *MapData = NULL;
static u_long MapSize = 0, MapOffset = 0;
static struct BootOptions *BootOptions = NULL;
static struct BootRecord *BootRecords = NULL;
static struct FileDef *Files = NULL;


static u_short MasterMode = FALSE;
static u_short Auto = FALSE;
static u_short Aux = FALSE;
static u_short Debug = FALSE;
static const char *Prompt = NULL;
static u_short MinKick = 0;
static u_short MaxKick = 65535;


    /*
     *	Linux/m68k Bootstrap Stuff
     */

static const struct BootRecord *DefaultBootRecord;
struct BootData BootData;


    /*
     *	Function Prototypes
     */

u_long Main(const struct LiloTable *lilotable);
static u_long GetDebugTag(void);
static const char *GetTagName(u_long tag);
static const struct TagRecord *GetNextTag(void);
static void UnexpectedTag(const struct TagRecord *tr);
static int EqualStrings(const char *s1, const char *s2);
static void ReadMapData(void);
static void FreeMapData(void);
static const u_long *FindVector(const char *path);
#ifdef V36_COMPATIBILITY
static u_short PeekQualifier36(void);
#endif /* V36_COMPATIBILITY */
static u_long ProbeConsole(void);
static void OpenConsole(void);
static void CloseConsole();
static long GetChar(void);
static long ReadLine(u_long mode);
static void SetTimeOut(u_long seconds);
static int TimedOut(void);
static void ClearTimeOut(void);
static void Sleep(u_long micros);
static int IsAvailable(const struct BootRecord *record);
static void GetBootOS(void);
static u_long BootLinux(struct BootData *boot);
static u_long BootNetBSD(struct BootData *boot);
static int Open(const char *path);
static int Seek(int fd, int offset);
static int Read(int fd, char *buf, int count);
static void Close(int fd);
static int FileSize(const char *path);
static int ModifyBootinfo(struct bootinfo *bi);


    /*
     *	Main Loader Routine
     *
     *	This routine should return FALSE in case of an error (which will cause
     *	a reboot), and TRUE if the user wants to boot AmigaOS.
     *	In all other cases it won't return :-)
     */

u_long Main(const struct LiloTable *lilotable)
{
    u_long res = FALSE;
    u_short interactive = FALSE;

    /* Boot Library */
    LiloTable = lilotable;

    /* Open Libraries */
    SysBase = *(struct ExecBase **)4;
    if (!(ExpansionBase =
	  (struct ExpansionBase *)OpenLibrary("expansion.library", LIB_VERS)))
	Alert(AN_LILO|AG_OpenLib|AO_ExpansionLib);
    if (!(GfxBase = (struct GfxBase *)OpenLibrary("graphics.library",
						  LIB_VERS)))
	Alert(AN_LILO|AG_OpenLib|AO_GraphicsLib);

    /* Open Timer Device */
    if (!(TimerPort = CreateMsgPort()))
	Alert(AN_LILO|AG_NoSignal);
    if (!(TimerRequest = CreateIORequest(TimerPort,
					 sizeof(struct timerequest))))
	Alert(AN_LILO|AG_NoMemory);
    if (OpenDevice("timer.device", UNIT_VBLANK,
		   (struct IORequest *)TimerRequest, 0))
	Alert(AN_LILO|AG_OpenDev|AO_TimerDev);

    /* Open Input Device */
    if (!(InputPort = CreateMsgPort()))
	Alert(AN_LILO|AG_NoSignal);
    if (!(InputRequest = CreateIORequest(InputPort, sizeof(struct IOStdReq))))
	Alert(AN_LILO|AG_NoMemory);
    if (OpenDevice("input.device", 0, (struct IORequest *)InputRequest, 0))
	Alert(AN_LILO|AG_OpenDev|AO_Unknown);
    InputBase = (struct InputBase *)InputRequest->io_Device;

    /* Configuration */
    ReadMapData();

    if (!Auto || ProbeConsole()) {
	OpenConsole();
	interactive = TRUE;
    }

    if (SysBase->Version < MinKick || SysBase->Version > MaxKick) {
	Printf("\nWarning: Kickstart version %ld is out of range %ld-%ld\n\n",
	       SysBase->Version, MinKick, MaxKick);
	if (!interactive) {
	    Puts("Autobooting AmigaOS...\n\n");
	    res = TRUE;
	    goto done;
	}
    }

    do {
	DefaultBootRecord = FindBootRecord(NULL);

	if (!DefaultBootRecord) {
	    Puts("\nPanic: No boot Operating System!\n");
	    break;
	}
	FillBootData(DefaultBootRecord, &BootData, TRUE);
	if (interactive)
	    GetBootOS();

	switch (BootData.OSType) {
	    case BOS_AMIGA:
		Puts("\nBooting AmigaOS...\n\n");
		res = TRUE;
		break;
	    case BOS_LINUX:
		res = BootLinux(&BootData);
		break;
	    case BOS_NETBSD:
		Puts("\nBooting NetBSD-Amiga...\n\n");
		res = BootNetBSD(&BootData);
		break;
	}
	if (res)
	    break;
	Puts("\nCouldn't boot the desired Operating System\n");
    } while (interactive);

done:
    CloseConsole();
    FreeMapData();

    /* Close Input Device */
    CloseDevice((struct IORequest *)InputRequest);
    DeleteIORequest(InputRequest);
    DeleteMsgPort(InputPort);

    /* Close Timer Device */
    CloseDevice((struct IORequest *)TimerRequest);
    DeleteIORequest(TimerRequest);
    DeleteMsgPort(TimerPort);

    /* Close Libraries */
    CloseLibrary((struct Library *)GfxBase);
    CloseLibrary((struct Library *)ExpansionBase);

    return(res);
}


    /*
     *	Get the Value of the Debug Tag (if it exists)
     */

static u_long GetDebugTag(void)
{
    const struct TagRecord *tr;
    u_long offset = 0;
    u_long debug = FALSE;

    while (offset+sizeof(struct TagRecord)+sizeof(u_long) <= MapSize) {
	tr = (struct TagRecord *)&MapData[offset];
	if (tr->Tag == TAG_DEBUG) {
	    debug = tr->Data[0];
	    break;
	}
	offset += sizeof(struct TagRecord)+tr->Size;
    }
    return(debug);
}


    /*
     *	Get the Name for a Tag
     */

static const char *GetTagName(u_long tag)
{
    const char *name = NULL;

    switch (tag) {
	case TAG_LILO:
	    name = "TAG_LILO";
	    break;
	case TAG_EOF:
	    name = "TAG_EOF";
	    break;
	case TAG_HEADER:
	    name = "TAG_HEADER";
	    break;
	case TAG_HEADER_END:
	    name = "TAG_HEADER_END";
	    break;
	case TAG_DEFAULT:
	    name = "TAG_DEFAULT";
	    break;
	case TAG_AUTO:
	    name = "TAG_AUTO";
	    break;
	case TAG_TIME_OUT:
	    name = "TAG_TIME_OUT";
	    break;
	case TAG_AUX:
	    name = "TAG_AUX";
	    break;
	case TAG_BAUD:
	    name = "TAG_BAUD";
	    break;
	case TAG_MASTER_PASSWORD:
	    name = "TAG_MASTER_PASSWORD";
	    break;
	case TAG_DEBUG:
	    name = "TAG_DEBUG";
	    break;
	case TAG_PROMPT:
	    name = "TAG_PROMPT";
	    break;
	case TAG_KICKRANGE:
	    name = "TAG_KICKRANGE";
	    break;
	case TAG_BOOT_RECORD:
	    name = "TAG_BOOT_RECORD";
	    break;
	case TAG_BOOT_RECORD_END:
	    name = "TAG_BOOT_RECORD_END";
	    break;
	case TAG_ALIAS:
	    name = "TAG_ALIAS";
	    break;
	case TAG_OS_TYPE:
	    name = "TAG_OS_TYPE";
	    break;
	case TAG_KERNEL:
	    name = "TAG_KERNEL";
	    break;
	case TAG_ARGUMENTS:
	    name = "TAG_ARGUMENTS";
	    break;
	case TAG_PASSWORD:
	    name = "TAG_PASSWORD";
	    break;
	case TAG_CHIP_RAM_SIZE:
	    name = "TAG_CHIP_RAM_SIZE";
	    break;
	case TAG_FAST_RAM_CHUNK:
	    name = "TAG_FAST_RAM_CHUNK";
	    break;
	case TAG_MODEL:
	    name = "TAG_MODEL";
	    break;
	case TAG_RAMDISK:
	    name = "TAG_RAMDISK";
	    break;
	case TAG_FILE_DEF:
	    name = "TAG_FILE_DEF";
	    break;
	case TAG_FILE_DEF_END:
	    name = "TAG_FILE_DEF_END";
	    break;
	case TAG_VECTOR:
	    name = "TAG_VECTOR";
	    break;
    }
    return(name);
}


    /*
     *	Get the Next Tag from the Map Data
     */

static const struct TagRecord *GetNextTag(void)
{
    static const struct TagRecord *tr = NULL;

    if (tr)
	MapOffset += sizeof(struct TagRecord)+tr->Size;
    if (MapOffset+sizeof(struct TagRecord) > MapSize)
	Alert(AN_LILO|AO_LiloMap);
    tr = (struct TagRecord *)&MapData[MapOffset];
    return(tr);
}


    /*
     *	Unexpected Tag
     */

static void UnexpectedTag(const struct TagRecord *tr)
{
    const char *name = GetTagName(tr->Tag);

    if (name)
	Printf("Unexpected tag `%s' at offset 0x%08lx\n", name, MapOffset);
    else
	Printf("Unknown tag 0x%08lx at offset 0x%08lx\n", tr->Tag, MapOffset);
}


    /*
     *	Check whether two strings exist and are equal
     */

static int EqualStrings(const char *s1, const char *s2)
{
    return(s1 && s2 && !strcmp(s1, s2) ? 1 : 0);
}


    /*
     *	Read the Map Data
     */

static void ReadMapData(void)
{
    const struct TagRecord *tr;
    struct BootRecord *record;
    struct BootRecord **p1 = &BootRecords;
    struct FileDef *file;
    struct FileDef **p2 = &Files;
    u_short isdefault = FALSE;
    int i;

    Lilo_Open(MapVectorData.Vector);
    MapSize = MapVectorData.Vector[0];
    if (!(MapData = (u_char *)AllocMem(MapSize, MEMF_PUBLIC|MEMF_CLEAR)))
	Alert(AN_LILO|AG_NoMemory);
    Lilo_Read(MapData, MapSize);
    Lilo_Close();
    if ((Debug = GetDebugTag()))
	OpenConsole();

    /* File Identification */
    tr = GetNextTag();
    if (tr->Tag != TAG_LILO)
	Alert(AN_LILO|AO_LiloMap);

    /* Boot Options */
    if (Debug)
	Puts("  [Boot Options]\n");
    tr = GetNextTag();
    if (tr->Tag != TAG_HEADER)
	Alert(AN_LILO|AO_LiloMap);
    if (!(BootOptions = (struct BootOptions *)AllocMem(sizeof(*BootOptions),
						       MEMF_PUBLIC|MEMF_CLEAR)))
	Alert(AN_LILO|AG_NoMemory);
    for (tr = GetNextTag(); tr->Tag != TAG_HEADER_END; tr = GetNextTag())
	switch (tr->Tag) {
	    case TAG_DEFAULT:
		BootOptions->Default = (const char *)tr->Data;
		break;
	    case TAG_AUTO:
		BootOptions->Auto = (const u_long *)tr->Data;
		break;
	    case TAG_TIME_OUT:
		BootOptions->TimeOut = (const u_long *)tr->Data;
		break;
	    case TAG_AUX:
		BootOptions->Aux = (const u_long *)tr->Data;
		break;
	    case TAG_BAUD:
		BootOptions->Baud = (const u_long *)tr->Data;
		break;
	    case TAG_MASTER_PASSWORD:
		BootOptions->MasterPassword = (const char *)tr->Data;
		break;
	    case TAG_DEBUG:
		BootOptions->Debug = (const u_long *)tr->Data;
		break;
	    case TAG_PROMPT:
		BootOptions->Prompt = (const char *)tr->Data;
		break;
	    default:
		UnexpectedTag(tr);
		break;;
	}

    if (BootOptions->Auto)
	Auto = *BootOptions->Auto;
    if (BootOptions->Aux)
	Aux = *BootOptions->Aux;
    if (BootOptions->Debug)
	Debug = *BootOptions->Debug;
    if (BootOptions->Baud) {
	u_long baud = *BootOptions->Baud;
	if (baud)
	    custom.serper = (5*SysBase->ex_EClockFrequency+baud/2)/baud-1;
    }
    if (!BootOptions->Default)
	isdefault = TRUE;
    if (BootOptions->KickRange) {
	MinKick = BootOptions->KickRange[0];
	MaxKick = BootOptions->KickRange[1];
    }
    Prompt = BootOptions->Prompt ? BootOptions->Prompt : "Boot image: ";

    /* Boot Records */
    if (Debug)
	Puts("  [Boot Records]\n");
    tr = GetNextTag();
    while (tr->Tag == TAG_BOOT_RECORD) {
	if (!(record = (struct BootRecord *)AllocMem(sizeof(*record),
						     MEMF_PUBLIC|MEMF_CLEAR)))
	    Alert(AN_LILO|AG_NoMemory);
	record->Label = (const char *)tr->Data;
	for (tr = GetNextTag(); tr->Tag != TAG_BOOT_RECORD_END;
	     tr = GetNextTag())
	    switch (tr->Tag) {
		case TAG_ALIAS:
		    record->Alias = (const u_char *)tr->Data;
		    break;
		case TAG_OS_TYPE:
		    record->OSType = (const u_long *)tr->Data;
		    break;
		case TAG_KERNEL:
		    record->Kernel = (const u_char *)tr->Data;
		    break;
		case TAG_ARGUMENTS:
		    record->Args = (const u_char *)tr->Data;
		    break;
		case TAG_PASSWORD:
		    record->Password = (const u_char *)tr->Data;
		    break;
		case TAG_CHIP_RAM_SIZE:
		    record->ChipSize = (const u_long *)tr->Data;
		    break;
		case TAG_FAST_RAM_CHUNK:
		    for (i = 0; i < arraysize(record->FastChunks); i++)
			if (!record->FastChunks[i])
			    break;
		    if (i < arraysize(record->FastChunks))
			record->FastChunks[i] = (const u_long *)tr->Data;
		    else
			UnexpectedTag(tr);
		    break;
		case TAG_MODEL:
		    record->Model = (const u_long *)tr->Data;
		    break;
		case TAG_RAMDISK:
		    record->Ramdisk = (const u_char *)tr->Data;
		    break;
		default:
		    UnexpectedTag(tr);
		    break;;
	    }

	if (EqualStrings(BootOptions->Default, record->Label) ||
	    EqualStrings(BootOptions->Default, record->Alias))
	    isdefault = TRUE;

	if (Debug) {
	    if (isdefault)
		Printf("\e[1m");
	    Printf("    %s\n", record->Label);
	    if (isdefault) {
		Printf("\e[0m");
		isdefault = FALSE;
	    }
	}

	*p1 = record;
	p1 = &record->Next;
	tr = GetNextTag();
    }

    /* File Definitions */
    if (Debug)
	Puts("  [File Definitions]\n");
    while (tr->Tag == TAG_FILE_DEF) {
	if (!(file = (struct FileDef *)AllocMem(sizeof(*file),
						MEMF_PUBLIC|MEMF_CLEAR)))
	    Alert(AN_LILO|AG_NoMemory);
	file->Path = (const char *)tr->Data;
	for (tr = GetNextTag(); tr->Tag != TAG_FILE_DEF_END; tr = GetNextTag())
	    switch (tr->Tag) {
		case TAG_VECTOR:
		    file->Vector = (const u_long *)tr->Data;
		    break;
		default:
		    UnexpectedTag(tr);
		    break;;
	    }
	if (Debug) {
	    Printf("    %s (", file->Path);
	    if (file->Vector)
		Printf("%ld bytes)\n", file->Vector[0]);
	    else
		Puts("not available)\n");
	}

	*p2 = file;
	p2 = &file->Next;
	tr = GetNextTag();
    }
    if (tr->Tag != TAG_EOF)
	UnexpectedTag(tr);
}


    /*
     *	Free the Map Data
     */

static void FreeMapData(void)
{
    struct BootRecord *record, *next;

    for (record = BootRecords; record; record = next) {
	next = record->Next;
	FreeMem(record, sizeof(struct BootRecord));
    }
    if (BootOptions)
	FreeMem(BootOptions, sizeof(struct BootOptions));
    if (MapData)
	FreeMem(MapData, MapSize);
}


    /*
     *	Find a Boot Record by Name
     */

const struct BootRecord *FindBootRecord(const char *name)
{
    const struct BootRecord *record;

    if (!name)
	name = BootOptions->Default;
    for (record = BootRecords; record; record = record->Next)
	if (EqualStrings(name, record->Label) ||
	    EqualStrings(name, record->Alias))
	    break;
    return(record);
}


    /*
     *  Find the File Vector for a File Path
     */

static const u_long *FindVector(const char *path)
{
    const struct FileDef *file;

    for (file = Files; file; file = file->Next)
	if (EqualStrings(path, file->Path))
	    break;
    return(file ? file->Vector : NULL);
}


#ifdef V36_COMPATIBILITY
    /*
     *	V36 (A3000 Beta ROM) compatible replacement for PeekQualifier()
     */

static u_short PeekQualifier36(void)
{
    u_short keys;

    /* Open Keyboard Device */

    if (!(KeyboardPort = CreateMsgPort()))
	Alert(AN_LILO|AG_NoSignal);
    if (!(KeyboardRequest = CreateIORequest(KeyboardPort,
					    sizeof(struct IOStdReq))))
	Alert(AN_LILO|AG_NoMemory);
    if (OpenDevice("keyboard.device", 0, (struct IORequest *)KeyboardRequest,
		   0))
	Alert(AN_LILO|AG_OpenDev|AO_KeyboardDev);
    KeyboardBase = (struct KeyboardBase *)KeyboardRequest->io_Device;

    KeyboardRequest->io_Command = KBD_READMATRIX;
    KeyboardRequest->io_Flags = IOF_QUICK;
    KeyboardRequest->io_Length = sizeof(KeyboardMatrix);
    KeyboardRequest->io_Data = KeyboardMatrix;
    DoIO((struct IORequest *)KeyboardRequest);

    keys = KeyboardMatrix[12];
    if (!(custom.potgor & 0x0100))
	keys |= IEQUALIFIER_MIDBUTTON;
    if (!(custom.potgor & 0x0400))
	keys |= IEQUALIFIER_RBUTTON;
    if (!(ciaa.pra & 0x40))
	keys |= IEQUALIFIER_LEFTBUTTON;

    /* Close Keyboard Device */

    CloseDevice((struct IORequest *)KeyboardRequest);
    DeleteIORequest(KeyboardRequest);
    DeleteMsgPort(KeyboardPort);

    return(keys);
}
#endif /* V36_COMPATIBILITY */


    /*
     *	Probe the Console
     */

static u_long ProbeConsole(void)
{
    u_long res;

#ifdef V36_COMPATIBILITY
    if (SysBase->Version >= 37)
#endif /* V36_COMPATIBILITY */
	res = PeekQualifier() ? TRUE : FALSE;
#ifdef V36_COMPATIBILITY
    else
	res = PeekQualifier36() ? TRUE : FALSE;
#endif /* V36_COMPATIBILITY */

    if (!res && Aux) {
	SetTimeOut(1);
	do
	    if (RawMayGetChar() != -1) {
		res = TRUE;
		while (RawMayGetChar() != -1);
		break;
	    }
	while (!TimedOut());
	ClearTimeOut();
    }
    return(res);
}


    /*
     *	Replacement for exec.library/Debug()
     *
     *	We don't want to meet a debugger when DEL is pressed on the terminal
     */

asm(".text\n"
ALIGN_STR "\n"
SYMBOL_NAME_STR(NewDebug) ":
	rts
");

extern void NewDebug(void);


    /*
     *	Open the Console
     */

static void OpenConsole(void)
{
    if (!ConsoleOpened) {
	if (!(IntuitionBase =
	      (struct IntuitionBase *)OpenLibrary("intuition.library",
						  LIB_VERS)))
	    Alert(AN_LILO|AG_OpenLib|AO_Intuition);
	NewScreen.Width = GfxBase->NormalDisplayColumns;
	NewScreen.Height = GfxBase->NormalDisplayRows;
	NewScreen.DefaultTitle = Title;
	if (!(Screen = OpenScreen(&NewScreen)))
	    Alert(AN_LILO|AN_OpenScreen);
	NewWindow.Width = NewScreen.Width;
	NewWindow.Height = NewScreen.Height;
	NewWindow.Screen = Screen;
	NewWindow.Title = Title;
	if (!(Window = OpenWindow(&NewWindow)))
	    Alert(AN_LILO|AN_OpenWindow);
	if (!(ConsolePort = CreateMsgPort()))
	    Alert(AN_LILO|AG_NoSignal);
	if (!(ConsoleRequest = CreateIORequest(ConsolePort,
					       sizeof(struct IOStdReq))))
	    Alert(AN_LILO|AG_NoMemory);
	ConsoleRequest->io_Data = Window;
	if (OpenDevice("console.device", CONU_SNIPMAP,
		       (struct IORequest *)ConsoleRequest, CONFLAG_DEFAULT))
	    Alert(AN_LILO|AG_OpenDev|AO_ConsoleDev);
	if (Aux)
	    OldDebug = SetFunction((struct Library *)SysBase, LVODebug,
				   NewDebug);
	ConsoleOpened = TRUE;
    }
}


    /*
     *	Close the Console
     */

static void CloseConsole()
{
    if (ConsoleOpened) {
	if (Aux)
	    SetFunction((struct Library *)SysBase, LVODebug, OldDebug);
	CloseDevice((struct IORequest *)ConsoleRequest);
	DeleteIORequest(ConsoleRequest);
	DeleteMsgPort(ConsolePort);
	CloseWindow(Window);
	CloseScreen(Screen);
	CloseLibrary((struct Library*)IntuitionBase);
	ConsoleOpened = FALSE;
    }
}


    /*
     *	Print a String to the Console
     */

void Puts(const char *str)
{
    if (ConsoleOpened) {
	ConsoleRequest->io_Command = CMD_WRITE;
	ConsoleRequest->io_Length = -1;
	ConsoleRequest->io_Data = (char *)str;
	SendIO((struct IORequest *)ConsoleRequest);
	if (Aux)
	    while (*str)
		RawPutChar(*str++);
	WaitIO((struct IORequest *)ConsoleRequest);
    }
}


    /*
     *	Put a Character to the Console
     */

void PutChar(char c)
{
    if (ConsoleOpened) {
	ConsoleRequest->io_Command = CMD_WRITE;
	ConsoleRequest->io_Length = 1;
	ConsoleRequest->io_Data = &c;
	SendIO((struct IORequest *)ConsoleRequest);
	if (Aux)
	    RawPutChar(c);
	WaitIO((struct IORequest *)ConsoleRequest);
    }
}


    /*
     *	Get a Character from the Console
     */

static long GetChar(void)
{
    long res = -1;
    char c;

    if (ConsoleOpened) {
	ConsoleRequest->io_Command = CMD_READ;
	ConsoleRequest->io_Length = 1;
	ConsoleRequest->io_Data = &c;
	SendIO((struct IORequest *)ConsoleRequest);
	do
	    if (Aux && ((res = RawMayGetChar()) != -1))
		break;
	    else if (CheckIO((struct IORequest *)ConsoleRequest)) {
		res = c;
		break;
	    }
	while (MasterMode || Debug || !TimedOut());
	AbortIO((struct IORequest *)ConsoleRequest);
	WaitIO((struct IORequest *)ConsoleRequest);
    }
    return(res);
}


    /*
     *	Format and Print a String to the Console
     */

asm(".text\n"
ALIGN_STR "\n"
SYMBOL_NAME_STR(PutCharInBuf) ":
	movel	a0,sp@-
	lea	pc@(" SYMBOL_NAME_STR(FmtBuffer) "+255),a0
	cmpl	a0,a3
	jne	1f
	clrb	a3@
	jra	2f
1:	moveb	d0,a3@+
2:	movel	sp@+,a0
	rts
");

extern void *PutCharInBuf(char c);

void Printf(const char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    RawDoFmt(fmt, args, (void (*)())PutCharInBuf, FmtBuffer);
    Puts(FmtBuffer);
    va_end(args);
}


    /*
     *	Read a Line from the Console
     */

static long ReadLine(u_long mode)
{
    long c;
    long length = 0;
    u_long eol = FALSE;

    while (!eol && ((c = GetChar()) != -1))
	switch (c) {
	    case '\r':
	    case '\n':
		InBuffer[length] = '\0';
		if ((mode == ECHO_NORMAL) || (mode == ECHO_DOT))
		    PutChar('\n');
		eol = TRUE;
		break;
	    case '\b':
	    case 0x7f:
		if (length > 0) {
		    length--;
		    if ((mode == ECHO_NORMAL) || (mode == ECHO_DOT))
			Puts("\b \b");
		}
		break;
	    case '\t':
		if (!MasterMode) {
		    strcpy(InBuffer, "help");
		    length = strlen("help");
		    switch (mode) {
			case ECHO_NORMAL:
			    Puts("help\n");
			    break;
			case ECHO_DOT:
			    Puts("++++\n");
			    break;
		    }
		    eol = TRUE;
		    break;
		}
	    default:
		if (length < sizeof(InBuffer)-1) {
		    InBuffer[length++] = c;
		    switch (mode) {
			case ECHO_NORMAL:
			    PutChar(c);
			    break;
			case ECHO_DOT:
			    PutChar('+');
			    break;
		    }
		}
		break;
	}
    return(eol ? length : -1);
}


    /*
     *  Read the Command Line
     */

char *ReadCommandLine(void)
{
    ReadLine(ECHO_NORMAL);
    return(InBuffer);
}


    /*
     *	Set the Time Out Value
     */

static void SetTimeOut(u_long seconds)
{
    if (!TimeOutSignalMask) {
	TimerRequest->io_Command = TR_ADDREQUEST;
	TimerRequest->tv_secs = seconds;
	TimerRequest->tv_micro = 0;
	SendIO((struct IORequest *)TimerRequest);
	TimeOutSignalMask = 1<<TimerPort->mp_SigBit;
    }
}


    /*
     *	Check for a Time Out
     */

static int TimedOut(void)
{
    int res = FALSE;

    if (TimeOutSignalMask && CheckIO((struct IORequest *)TimerRequest))
	res = TRUE;

    return(res);
}


    /*
     *	Clear Time Out
     */

static void ClearTimeOut(void)
{
    if (TimeOutSignalMask) {
	AbortIO((struct IORequest *)TimerRequest);
	WaitIO((struct IORequest *)TimerRequest);
	TimeOutSignalMask = 0;
    }
}


    /*
     *	Sleep for a Specified Period
     */

static void Sleep(u_long micros)
{
    TimerRequest->io_Command = TR_ADDREQUEST;
    TimerRequest->io_Flags = IOF_QUICK;
    TimerRequest->tv_secs = micros/1000000;
    TimerRequest->tv_micro = micros%1000000;
    DoIO((struct IORequest *)TimerRequest);
}


    /*
     *  Check whether a Boot Image is available
     */

static int IsAvailable(const struct BootRecord *record)
{
    u_long type = record->OSType ? *record->OSType : BOS_LINUX;
    int res = TRUE;

    if (type != BOS_AMIGA) {
	if (!FindVector(record->Kernel))
	    res = FALSE;
	if (record->Ramdisk && !FindVector(record->Ramdisk))
	    res = FALSE;
    }
    return(res);
}


    /*
     *  List all available Boot Records
     */

void ListRecords(void)
{
    const struct BootRecord *record;

    Puts("\nBoot images:\n");
    for (record = BootRecords; record; record = record->Next) {
	if (record == DefaultBootRecord)
	    Puts("\e[1m");
	Printf("    %-31.31s (", record->Label);
	if (record->Alias)
	    Printf("`%s', ", record->Alias);
	switch (record->OSType ? *record->OSType : BOS_LINUX) {
	    case BOS_AMIGA:
		Puts("AmigaOS");
		break;
	    case BOS_LINUX:
		Puts("Linux/m68k");
		break;
	    case BOS_NETBSD:
		Puts("NetBSD-Amiga");
		break;
	    default:
		Puts("Unknown");
		break;
	}
	if (record->Password)
	    Puts(", Restricted");
	if (!IsAvailable(record))
	    Puts(", N/A");
	Puts(")\n");
	if (record == DefaultBootRecord)
	    Puts("\e[0m");
    }
}


    /*
     *  List all available Files
     */

void ListFiles(void)
{
    const struct FileDef *file;

    Puts("\nFiles:\n");
    for (file = Files; file; file = file->Next) {
	Printf("    %s (", file->Path);
	if (file->Vector)
	    Printf("%ld bytes", file->Vector[0]);
	else
	    Puts("N/A");
	Puts(")\n");
    }
}


    /*
     *	Ask the User for a Boot Operating System
     */

static void GetBootOS(void)
{
    const struct BootRecord *record;

    if (BootOptions->TimeOut ? *BootOptions->TimeOut : 20)
	SetTimeOut(*BootOptions->TimeOut);

    while (1) {
	PutChar('\n');
	Puts(Prompt);
	if ((ReadLine(ECHO_NORMAL)) < 1)
	    break;
	if (!strcmp(InBuffer, "help") || !strcmp(InBuffer, "?")) {
	    Printf("\n%s\nValid commands: su, help, ?\n", LiloVersion);
	    ListRecords();
	    PutChar('\n');
	} else if (!strcmp(InBuffer, "su")) {
	    if (BootOptions->MasterPassword) {
		Puts("\nPassword: ");
		if (((ReadLine(ECHO_DOT)) < 1) ||
		     strcmp(InBuffer, BootOptions->MasterPassword)) {
		    Puts("\nIncorrect password.\n");
		    continue;
		}
	    }
	    MasterMode = TRUE;
	    Monitor();
	    break;
	} else {
	    if (!(record = FindBootRecord(InBuffer))) {
		Printf("\nNonexistent boot image `%s'.\n", InBuffer);
		continue;
	    }
	    if (!IsAvailable(record)) {
		Printf("\nBoot image `%s' is not available.\n", InBuffer);
		continue;
	    } else if (record->Password) {
		Puts("\n\nPassword: ");
		if (((ReadLine(ECHO_DOT)) < 1) ||
		    strcmp(InBuffer, record->Password)) {
		    Puts("\nIncorrect password.\n");
		    continue;
		}
	    }
	    FillBootData(record, &BootData, FALSE);
	    break;
	}
    }
    ClearTimeOut();
}


    /*
     *  Fill the Boot Data according to a Boot Record
     */

void FillBootData(const struct BootRecord *record, struct BootData *boot,
		  int autoboot)
{
    int i;

    boot->OSType = record->OSType ? *record->OSType : BOS_LINUX;
    if (record->Kernel) {
	strncpy(boot->Kernel, record->Kernel, PATH_SIZE);
	boot->Kernel[PATH_SIZE-1] = '\0';
    } else
	boot->Kernel[0] = '\0';
    if (record->Ramdisk) {
	strncpy(boot->Ramdisk, record->Ramdisk, PATH_SIZE);
	boot->Ramdisk[PATH_SIZE-1] = '\0';
    } else
	boot->Ramdisk[0] = '\0';
    if (record->Args) {
	strncpy(boot->Args, record->Args, CL_SIZE);
	strncat(boot->Args, " BOOT_IMAGE=", CL_SIZE);
    } else
	boot->Args[0] = '\0';
    strncat(boot->Args, record->Label, CL_SIZE);
    boot->Args[CL_SIZE-1] = '\0';
    if (autoboot)
	strncat(boot->Args, " auto", CL_SIZE);
    boot->Debug = Debug;
    boot->Model = record->Model ? *record->Model : AMI_UNKNOWN;
    boot->ChipSize = record->ChipSize ? *record->ChipSize : 0;
    boot->NumMemory = 0;
    for (i = 0; i < NUM_MEMINFO; i++)
	if (record->FastChunks[i]) {
	    boot->FastChunks[i].Address = record->FastChunks[i][0];
	    boot->FastChunks[i].Size = record->FastChunks[i][1];
	    boot->NumMemory++;
	}
}


    /*
     *  Boot the Linux/m68k Operating System
     */

static u_long BootLinux(struct BootData *boot)
{
    struct linuxboot_args args;

    Puts("\nBooting Linux/m68k...\n\n");
    args.kernelname = boot->Kernel[0] ? boot->Kernel : NULL;
    args.ramdiskname = boot->Ramdisk[0] ? boot->Ramdisk : NULL;
    args.commandline = boot->Args;
    args.debugflag = boot->Debug;
    args.keep_video = 0;
    args.reset_boards = 0;
    args.puts = Puts;
    args.getchar = GetChar;
    args.putchar = PutChar;
    args.printf = Printf;
    args.open = Open;
    args.seek = Seek;
    args.read = Read;
    args.close = Close;
    args.filesize = FileSize;
    args.sleep = Sleep;
    args.modify_bootinfo = ModifyBootinfo;

    /* Do The Right Stuff */
    return(linuxboot(&args));
}


    /*
     *  Boot the NetBSD-Amiga Operating System
     */

static u_long BootNetBSD(struct BootData *boot)
{
    Puts("Not yet implemented :-)\n");
    return(FALSE);
}


    /*
     *	Functions needed by LinuxBoot
     *
     *	These functions are not complete implementations. They do only what's
     *	really necessary.
     */

static int OpenFiles = 0;
static int FileOffset;
const static u_long *CurrentVector;

static int Open(const char *path)
{
    if (OpenFiles) {
	Puts("Internal error: Only one file can be open\n");
	return(-1);
    }
    if (!(CurrentVector = FindVector(path)))
	return(-1);
    Lilo_Open(CurrentVector);
    FileOffset = 0;
    OpenFiles++;
    return(1);
}


static int Seek(int fd, int offset)
{
    if (offset > CurrentVector[0])
	return(-1);
    Lilo_Seek(offset);
    FileOffset = offset;
    return(offset);
}


static int Read(int fd, char *buf, int count)
{
    if (count > CurrentVector[0]-FileOffset)
	count = CurrentVector[0]-FileOffset;
    Lilo_Read(buf, count);
    FileOffset += count;
    return(count);
}


static void Close(int fd)
{
    OpenFiles--;
    Lilo_Close();
}


static int FileSize(const char *path)
{
    const u_long *vector;

    if (!(vector = FindVector(path)))
	return(-1);
    return(vector[0]);
}


    /*
     *	Modify the Bootinfo, if necessary
     */

static int ModifyBootinfo(struct bootinfo *bi)
{
    int i, j;

    if (Debug)
	Puts("Modifying bootinfo structure\n");

    if (BootData.ChipSize) {
	bi->bi_amiga.chip_size = BootData.ChipSize;
	if (Debug)
	    Printf("Chip RAM: %10ld bytes\n", bi->bi_amiga.chip_size);
    }

    if (BootData.NumMemory) {
	for (i = 0, j = 0; i < BootData.NumMemory; i++)
	    if (BootData.FastChunks[i].Size) {
		bi->memory[j].addr = BootData.FastChunks[i].Address;
		bi->memory[j].size = BootData.FastChunks[i].Size;
		if (Debug)
		    Printf("Fast RAM: %10ld bytes at 0x%08lx\n",
			   bi->memory[j].addr, bi->memory[j].size);
		j++;
	    }
	if (j)
	    bi->num_memory = j;
    }

    if (BootData.Model != AMI_UNKNOWN) {
	bi->bi_amiga.model = BootData.Model;
	if (Debug)
	    Printf("Model   : #%ld\n", bi->bi_amiga.model);
    }

    return(TRUE);
}


    /*
     *  Integer Math
     *
     *  Libnix uses utility.library if these aren't present
     */

asm("
		.globl	___umodsi3
		.globl	___udivsi3
		.globl	___udivsi4
		.globl	___mulsi3

| D0.L = D0.L % D1.L unsigned

___umodsi3:	moveml	sp@(4:w),d0/d1
		jbsr	___udivsi4
		movel	d1,d0
		rts

| D0.L = D0.L / D1.L unsigned

___udivsi3:	moveml	sp@(4:w),d0/d1
___udivsi4:	movel	d3,sp@-
		movel	d2,sp@-
		movel	d1,d3
		swap	d1
		tstw	d1
		jne	1f
		movew	d0,d2
		clrw	d0
		swap	d0
		divu	d3,d0
		movel	d0,d1
		swap	d0
		movew	d2,d1
		divu	d3,d1
		movew	d1,d0
		clrw	d1
		swap	d1
		jra	4f
1:		movel	d0,d1
		swap	d0
		clrw	d0
		clrw	d1
		swap	d1
		moveq	#16-1,d2
2:		addl	d0,d0
		addxl	d1,d1
		cmpl	d1,d3
		jhi	3f
		subl	d3,d1
		addqw	#1,d0
3:		dbra	d2,2b
4:		movel	sp@+,d2
		movel	sp@+,d3
		rts

| D0 = D0 * D1

___mulsi3:	moveml	sp@(4:w),d0/d1
		movel	d3,sp@-
		movel	d2,sp@-
		movew	d1,d2
		mulu	d0,d2
		movel	d1,d3
		swap	d3
		mulu	d0,d3
		swap	d3
		clrw	d3
		addl	d3,d2
		swap	d0
		mulu	d1,d0
		swap	d0
		clrw	d0
		addl	d2,d0
		movel	sp@+,d2
		movel	sp@+,d3
		rts
");
