/*
 *  Amiga Linux/m68k Loader
 *
 *   Copyright 1995 by Geert Uytterhoeven
 *
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file COPYING in the main directory of the Linux
 *  distribution for more details.
 */


#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include "lilo.h"
#include "loader_proto.h"


    /*
     *  Function Prototypes
     */

void Monitor(void);
static int PartStrCaseCmp(const char *s1, const char *s2);
static void CreateArgs(char *cmdline, int *argcp, const char **argvp[]);
static void DeleteArgs(int argc, const char *argv[]);
static int ParseCommandLine(char *line);
static void Moni_help(void);
static void Moni_version(void);
static void Moni_list(int argc, const char *argv[]);
static int Moni_boot(int argc, const char *argv[]);
static int Moni_use(int argc, const char *argv[]);
static void Moni_info(int argc, const char *argv[]);
static void Moni_set(int argc, const char *argv[]);
static void Moni_set_type(int argc, const char *argv[]);
static void Moni_set_kernel(int argc, const char *argv[]);
static void Moni_set_ramdisk(int argc, const char *argv[]);
static void Moni_set_args(int argc, const char *argv[]);
static void Moni_set_debug(int argc, const char *argv[]);
static void Moni_set_model(int argc, const char *argv[]);
static void Moni_set_chip(int argc, const char *argv[]);
static void Moni_set_nummem(int argc, const char *argv[]);
static void Moni_set_mem(int argc, const char *argv[]);
static void DumpInfo(const struct BootData *data);


    /*
     *  The Boot Monitor
     */

void Monitor(void)
{
    char *line;

    Puts("\nEntering Lilo Boot Monitor.\n");
    do {
	Puts("lilo> ");
	line = ReadCommandLine();
    } while (!ParseCommandLine(line));
}


    /*
     *  Split a line into arguments
     */

static void CreateArgs(char *cmdline, int *argcp, const char **argvp[])
{
    int argc = 0, i;
    char *p, *q;
    const char **argv = NULL;

    *argcp = 0;
    p = cmdline;
    for (argc = 0;; argc++) {
	while (*p == ' ' || *p == '\t')
	    p++;
	if (!*p)
	    break;
	if (*p == '"') {
	    for (p++; *p && *p != '"'; p++)
		if (*p == '\\') {
		    p++;
		    if (!*p)
			break;
		}
	    if (*p)
		p++;
	} else
	    while (*p && *p != ' ' && *p != '\t')
		p++;
    }
    if (argc) {
	if (!(argv = AllocMem((argc+1)*sizeof(char *), MEMF_CLEAR)))
	    return;
	p = cmdline;
	for (i = 0; i < argc; i++) {
	    while (*p == ' ' || *p == '\t')
		p++;
	    if (!*p)
		break;
	    if (*p == '"') {
		q = p++;
		argv[i] = q;
		while (*p && *p != '"') {
		    if (*p == '\\') {
			p++;
			if (!*p)
			    break;
		    }
		    *q++ = *p++;
		}
	    } else {
		q = p;
		argv[i] = q;
		while (*p && *p != ' ' && *p != '\t') {
		    if (*p == '\\') {
			p++;
			if (!*p)
			    break;
		    }
		    *q++ = *p++;
		}
	    }
	    *q = '\0';
	    p++;
	}
    }
    *argcp = argc;
    *argvp = argv;
}


static void DeleteArgs(int argc, const char *argv[])
{
    if (argc)
	FreeMem(argv, (argc+1)*sizeof(char *));
}


    /*
     *  Partial strcasecmp() (s1 may be an abbreviation for s2)
     */

static int PartStrCaseCmp(const char *s1, const char *s2)
{
    char c1, c2;

    while (*s1) {
	c1 = *s1++;
	if (c1 >= 'A' && c1 <= 'Z')
	    c1 += ('a'-'A');
	c2 = *s2++;
	if (c2 >= 'A' && c2 <= 'Z')
	    c2 += ('a'-'A');
	if (!c2 || c1 != c2)
	    return(0);
    }
    return(1);
}


    /*
     *  Parse the Monitor Command Line
     */

static int ParseCommandLine(char *line)
{
    int argc;
    const char **argv;
    int do_exit = 0;

    CreateArgs(line, &argc, &argv);
    if (argc) {
	if (PartStrCaseCmp(argv[0], "help") || PartStrCaseCmp(argv[0], "?"))
	    Moni_help();
	else if (PartStrCaseCmp(argv[0], "version"))
	    Moni_version();
	else if (PartStrCaseCmp(argv[0], "list"))
	    Moni_list(argc-1, argv+1);
	else if (PartStrCaseCmp(argv[0], "boot"))
	    do_exit = Moni_boot(argc-1, argv+1);
	else if (PartStrCaseCmp(argv[0], "use"))
	    Moni_use(argc-1, argv+1);
	else if (PartStrCaseCmp(argv[0], "info"))
	    Moni_info(argc-1, argv+1);
	else if (PartStrCaseCmp(argv[0], "set"))
	    Moni_set(argc-1, argv+1);
	else
	    Puts("Unknown command\n");
    }
    DeleteArgs(argc, argv);
    return(do_exit);
}


static void Moni_help(void)
{
    Puts("\nLilo Boot Monitor\n\n"
	 "   help, ?                 Display this help\n"
	 "   version                 Version information\n"
	 "   list [images|files]     List all boot records and files\n"
	 "   boot [<label>]          Boot an Operating System\n"
	 "   use [<label>]           Use the specified boot record\n"
	 "   info [<label>]          Get information about a boot record\n"
	 "   set [<var>] [<value>]   Set boot variables\n\n");
}


static void Moni_version(void)
{
    Printf("%s\n", LiloVersion);
}


static void Moni_list(int argc, const char *argv[])
{
    if (!argc) {
	ListRecords();
	ListFiles();
	PutChar('\n');
    } else if (PartStrCaseCmp(argv[0], "images")) {
	ListRecords();
	PutChar('\n');
    } else if (PartStrCaseCmp(argv[0], "files")) {
	ListFiles();
	PutChar('\n');
    } else
	Printf("Unknown list type\n");
}


static int Moni_boot(int argc, const char *argv[])
{
    return(argc ? Moni_use(argc, argv) : 1);
}


static int Moni_use(int argc, const char *argv[])
{
    const struct BootRecord *record;
    int res = 0;
    const char *label = argc ? argv[0] : NULL;

    if ((record = FindBootRecord(label))) {
	FillBootData(record, &BootData, FALSE);
	res = 1;
    } else
	Puts("Nonexistent boot image\n");
    return(res);
}


static void Moni_info(int argc, const char *argv[])
{
    const struct BootRecord *record;
    struct BootData data;

    if (!argc)
	DumpInfo(&BootData);
    else if ((record = FindBootRecord(argv[0]))) {
	FillBootData(record, &data, FALSE);
	DumpInfo(&data);
    } else
	Puts("Nonexistent boot image\n");
}


static void Moni_set(int argc, const char *argv[])
{
    if (!argc)
	DumpInfo(&BootData);
    else if (PartStrCaseCmp(argv[0], "type"))
	Moni_set_type(argc-1, argv+1);
    else if (PartStrCaseCmp(argv[0], "kernel"))
	Moni_set_kernel(argc-1, argv+1);
    else if (PartStrCaseCmp(argv[0], "ramdisk"))
	Moni_set_ramdisk(argc-1, argv+1);
    else if (PartStrCaseCmp(argv[0], "args"))
	Moni_set_args(argc-1, argv+1);
    else if (PartStrCaseCmp(argv[0], "debug"))
	Moni_set_debug(argc-1, argv+1);
    else if (PartStrCaseCmp(argv[0], "model"))
	Moni_set_model(argc-1, argv+1);
    else if (PartStrCaseCmp(argv[0], "chip"))
	Moni_set_chip(argc-1, argv+1);
    else if (PartStrCaseCmp(argv[0], "nummem"))
	Moni_set_nummem(argc-1, argv+1);
    else if (PartStrCaseCmp(argv[0], "mem"))
	Moni_set_mem(argc-1, argv+1);
    else
	Printf("Unknown variable\n");
}


static void Moni_set_type(int argc, const char *argv[])
{
    if (!argc)
	Puts("You should specify an Operating System type\n");
    if (PartStrCaseCmp(argv[0], "amiga"))
	BootData.OSType = BOS_AMIGA;
    else if (PartStrCaseCmp(argv[0], "linux"))
	BootData.OSType = BOS_LINUX;
    else if (PartStrCaseCmp(argv[0], "netbsd"))
	BootData.OSType = BOS_NETBSD;
    else
	Puts("Unknown Operating System type\n");
}


static void Moni_set_kernel(int argc, const char *argv[])
{
    if (!argc)
	BootData.Kernel[0] = '\0';
    else {
	strncpy(BootData.Kernel, argv[0], PATH_SIZE);
	BootData.Kernel[PATH_SIZE-1] = '\0';
    }
}


static void Moni_set_ramdisk(int argc, const char *argv[])
{
    if (!argc)
	BootData.Ramdisk[0] = '\0';
    else {
	strncpy(BootData.Ramdisk, argv[0], PATH_SIZE);
	BootData.Ramdisk[PATH_SIZE-1] = '\0';
    }
}


static void Moni_set_args(int argc, const char *argv[])
{
    int i;

    BootData.Args[0] = '\0';
    for (i = 0; argc; argc--)
	if ((i+strlen(*argv)+1) < CL_SIZE) {
	    i += strlen(*argv)+1;
	    if (BootData.Args[0])
		strcat(BootData.Args, " ");
	    strcat(BootData.Args, *argv++);
	}
}


static void Moni_set_debug(int argc, const char *argv[])
{
    if (!argc || PartStrCaseCmp(argv[0], "TRUE"))
	BootData.Debug = TRUE;
    else if (PartStrCaseCmp(argv[0], "FALSE"))
	BootData.Debug = FALSE;
    else
	Printf("Unknown boolean value\n");
}


static void Moni_set_model(int argc, const char *argv[])
{
    if (!argc || PartStrCaseCmp(argv[0], "auto"))
	BootData.Model = AMI_UNKNOWN;
    else if (PartStrCaseCmp(argv[0], "A500"))
	BootData.Model = AMI_500;
    else if (PartStrCaseCmp(argv[0], "A500+"))
	BootData.Model = AMI_500PLUS;
    else if (PartStrCaseCmp(argv[0], "A600"))
	BootData.Model = AMI_600;
    else if (PartStrCaseCmp(argv[0], "A1000"))
	BootData.Model = AMI_1000;
    else if (PartStrCaseCmp(argv[0], "A1200"))
	BootData.Model = AMI_1200;
    else if (PartStrCaseCmp(argv[0], "A2000"))
	BootData.Model = AMI_2000;
    else if (PartStrCaseCmp(argv[0], "A2500"))
	BootData.Model = AMI_2500;
    else if (PartStrCaseCmp(argv[0], "A3000"))
	BootData.Model = AMI_3000;
    else if (PartStrCaseCmp(argv[0], "A3000T"))
	BootData.Model = AMI_3000T;
    else if (PartStrCaseCmp(argv[0], "A3000+"))
	BootData.Model = AMI_3000PLUS;
    else if (PartStrCaseCmp(argv[0], "A4000"))
	BootData.Model = AMI_4000;
    else if (PartStrCaseCmp(argv[0], "A4000T"))
	BootData.Model = AMI_4000T;
    else if (PartStrCaseCmp(argv[0], "CDTV"))
	BootData.Model = AMI_CDTV;
    else if (PartStrCaseCmp(argv[0], "CD32"))
	BootData.Model = AMI_CD32;
    else if (PartStrCaseCmp(argv[0], "Draco"))
	BootData.Model = AMI_DRACO;
    else
	BootData.Model = strtoul(argv[0], NULL, 0);
}


static void Moni_set_chip(int argc, const char *argv[])
{
    if (!argc)
	Puts("You should specify the Chip RAM size\n");
    else if (PartStrCaseCmp(argv[0], "auto"))
	BootData.ChipSize = 0;
    else
	BootData.ChipSize = strtoul(argv[0], NULL, 0);
}


static void Moni_set_nummem(int argc, const char *argv[])
{
    u_long i;

    if (!argc || PartStrCaseCmp(argv[0], "auto"))
	BootData.NumMemory = 0;
    else {
	i = strtoul(argv[0], NULL, 0);
	if (!i)
	    BootData.NumMemory = 0;
	else if (i > NUM_MEMINFO)
	    Puts("Invalid number of memory blocks\n");
	else
	    BootData.NumMemory = i;
    }
}


static void Moni_set_mem(int argc, const char *argv[])
{
    u_long i;

    if (argc < 3)
	Puts("You should specify a memory block number, address and size\n");
    else {
	i = strtoul(argv[0], NULL, 0);
	if (i < 1 || i > BootData.NumMemory)
	    Puts("Invalid memory block number\n");
	else {
	    BootData.FastChunks[i-1].Address = strtoul(argv[1], NULL, 0);
	    BootData.FastChunks[i-1].Size = strtoul(argv[2], NULL, 0);
	}
    }
}


static void DumpInfo(const struct BootData *data)
{
    int i;
    const char *p;

    PutChar('\n');
    switch (data->OSType) {
	case BOS_AMIGA:
	    p = "amiga (AmigaOS)";
	    break;
	case BOS_LINUX:
	    p = "linux (Linux/m68k)";
	    break;
	case BOS_NETBSD:
	    p = "netbsd (NetBSD-Amiga)";
	    break;
	default:
	    p = NULL;
	    break;
    }
    if (p)
	Printf("[type]    OS Type:          %s\n", p);
    else
	Printf("[type]    OS Type:          Unknown (%ld)\n", data->OSType);
    Printf("[kernel]  Kernel:           %s\n", data->Kernel);
    Printf("[ramdisk] Ramdisk:          %s\n", data->Ramdisk);
    Printf("[args]    Arguments:        %s\n", data->Args);
    Printf("[debug]   Debug Mode:       %s\n", data->Debug ? "TRUE" : "FALSE");
    switch (data->Model) {
	case AMI_UNKNOWN:
	    p = "auto";
	    break;
	case AMI_500:
	    p = "A500";
	    break;
	case AMI_500PLUS:
	    p = "A500+";
	    break;
	case AMI_600:
	    p = "A600";
	    break;
	case AMI_1000:
	    p = "A1000";
	    break;
	case AMI_1200:
	    p = "A1200";
	    break;
	case AMI_2000:
	    p = "A2000";
	    break;
	case AMI_2500:
	    p = "A2500";
	    break;
	case AMI_3000:
	    p = "A3000";
	    break;
	case AMI_3000T:
	    p = "A3000T";
	    break;
	case AMI_3000PLUS:
	    p = "A3000+";
	    break;
	case AMI_4000:
	    p = "A4000";
	    break;
	case AMI_4000T:
	    p = "A4000T";
	    break;
	case AMI_CDTV:
	    p = "CDTV";
	    break;
	case AMI_CD32:
	    p = "CD32";
	    break;
	case AMI_DRACO:
	    p = "Draco";
	    break;
	default:
	    p = NULL;
	    break;
    }
    if (p)
	Printf("[model]   Amiga Model:      %s\n", p);
    else
	Printf("[model]   Amiga Model:      Unknown (%ld)\n", data->Model);
    if (!data->ChipSize)
	Puts("[chip]    Chip RAM Size:    auto\n");
    else
	Printf("[chip]    Chip RAM Size:    0x%08lx (%ld)\n", data->ChipSize,
	   data->ChipSize);
    if (!data->NumMemory)
	Puts("[nummem]  Memory Blocks:    auto\n");
    else {
	Printf("[nummem]  Memory Blocks:    %ld\n", data->NumMemory);
	for (i = 0; i < data->NumMemory; i++)
	    Printf("[mem %2ld]  0x%08lx: 0x%08lx bytes\n", i+1,
		   data->FastChunks[i].Address, data->FastChunks[i].Size);
    }
    PutChar('\n');
}
